try:
    from loguru import logger
except ImportError: # pragma: no cover
    import logging
    logger = logging.getLogger(__name__)
import asyncio
import socket
import argparse
from typing import Dict, Tuple, Set, Optional, Sequence

from pydispatch import Dispatcher, Property, DictProperty, ListProperty

from tslumd import Message, Display, TallyColor, TallyType, Tally
from tslumd.utils import logger_catch

Client = Tuple[str, int] #: A network client as a tuple of ``(address, port)``

__all__ = ('Client', 'UmdSender')

class UmdProtocol(asyncio.DatagramProtocol):
    def __init__(self, sender: 'UmdSender'):
        self.sender = sender
    def connection_made(self, transport):
        logger.debug(f'transport={transport}')
        self.transport = transport
        self.sender.connected_evt.set()
    def datagram_received(self, data, addr): # pragma: no cover
        pass

class UmdSender(Dispatcher):
    """Send UMD Messages

    Messages are sent immediately when a change is made to any of the
    :class:`~.Tally` objects in :attr:`tallies`. These can be added by using
    the :meth:`add_tally` method.

    Alternatively, the :meth:`set_tally_color` and :meth:`set_tally_text` methods
    may be used.

    Arguments:
        clients: Intitial value for :attr:`clients`
    """

    tallies: Dict[int, Tally]
    """Mapping of :class:`Tally` objects using the :attr:`~Tally.index` as keys

    Note:
        This should not be altered directly. Use :meth:`add_tally` instead
    """

    running: bool
    """``True`` if the client / server are running
    """

    loop: asyncio.BaseEventLoop
    """The :class:`asyncio.BaseEventLoop` associated with the instance"""

    tx_interval: float = .3
    """Interval to send tally messages, regardless of state changes
    """

    clients: Set[Client]
    """Set of :data:`clients <Client>` to send messages to
    """

    def __init__(self, clients: Optional[Set[Client]] = None):
        self.clients = set()
        if clients is not None:
            for client in clients:
                self.clients.add(client)
        self.tallies = {}
        self.running = False
        self.loop = asyncio.get_event_loop()
        self.update_queue = asyncio.Queue()
        self.update_task = None
        self.tx_task = None
        self.connected_evt = asyncio.Event()

    async def open(self):
        """Open connections and begin data transmission
        """
        if self.running:
            return
        self.connected_evt.clear()
        logger.debug('UmdSender.open()')
        self.running = True
        self.transport, self.protocol = await self.loop.create_datagram_endpoint(
            lambda: UmdProtocol(self),
            family=socket.AF_INET,
        )
        self.tx_task = asyncio.create_task(self.tx_loop())
        logger.info('UmdSender running')

    async def close(self):
        """Stop sending to clients and close connections
        """
        if not self.running:
            return
        logger.debug('UmdSender.close()')
        self.running = False
        await self.update_queue.put(False)
        await self.tx_task
        self.tx_task = None
        self.transport.close()
        logger.info('UmdSender closed')

    def add_tally(self, index_: int, **kwargs) -> Tally:
        """Create a :class:`~.Tally` object and add it to :attr:`tallies`

        Arguments:
            index_: The tally :attr:`~.Tally.index`
            **kwargs: Keyword arguments passed to create the tally instance

        Raises:
            KeyError: If the given ``index_`` already exists
        """
        if index_ in self.tallies:
            raise KeyError(f'Tally exists for index {index_}')
        tally = Tally(index_, **kwargs)
        self.tallies[index_] = tally
        tally.bind_async(self.loop, on_update=self.on_tally_updated)
        logger.debug(f'new tally: {tally}')
        return tally

    def set_tally_color(self, index_: int, tally_type: TallyType, color: TallyColor):
        """Set the tally color for the given index and tally type

        Arguments:
            index_: The tally :attr:`~.Tally.index`
            tally_type: A member of :class:`~.common.TallyType` specifying the
                tally lamp within the display
            color: The member of :class:`~.common.TallyColor` to set
        """
        if tally_type == TallyType.no_tally:
            raise ValueError()
        if index_ not in self.tallies:
            tally = self.add_tally(index_)
        else:
            tally = self.tallies[index_]
        attr = tally_type.name
        setattr(tally, attr, color)

    def set_tally_text(self, index_: int, text: str):
        """Set the tally text for the given index

        Arguments:
            index_: The tally :attr:`~.Tally.index`
            text: The :attr:`~.Tally.text` to set
        """
        if index_ not in self.tallies:
            tally = self.add_tally(index_)
        else:
            tally = self.tallies[index_]
        tally.text = text

    async def on_tally_updated(self, tally: Tally, props_changed: Sequence[str], **kwargs):
        if self.running:
            logger.debug(f'tally update: {tally}')
            await self.update_queue.put(tally.index)

    @logger_catch
    async def tx_loop(self):
        async def get_queue_item(timeout):
            try:
                item = await asyncio.wait_for(self.update_queue.get(), timeout)
            except asyncio.TimeoutError:
                item = None
            return item

        await self.connected_evt.wait()

        while self.running:
            item = await get_queue_item(self.tx_interval)
            if item is False:
                self.update_queue.task_done()
                break
            elif item is None:
                await self.send_full_update()
            else:
                indices = set()
                indices.add(item)
                self.update_queue.task_done()
                while not self.update_queue.empty():
                    try:
                        item = self.update_queue.get_nowait()
                    except asyncio.QueueEmpty:
                        break
                    if item is False:
                        self.update_queue.task_done()
                        return
                    indices.add(item)
                    self.update_queue.task_done()
                msg = self._build_message()
                tallies = {i:self.tallies[i] for i in indices}
                for key in sorted(tallies.keys()):
                    tally = tallies[key]
                    msg.displays.append(Display.from_tally(tally))
                await self.send_message(msg)

    async def send_message(self, msg: Message):
        data = msg.build_message()
        for client in self.clients:
            self.transport.sendto(data, client)

    async def send_full_update(self):
        msg = self._build_message()
        for tally in self.tallies.values():
            disp = Display.from_tally(tally)
            msg.displays.append(disp)
        await self.send_message(msg)

    def _build_message(self) -> Message:
        return Message()

    async def __aenter__(self):
        await self.open()
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        await self.close()


class ClientArgAction(argparse._AppendAction):
    _default_help = ' '.join([
        'Client(s) to send UMD messages to formatted as "<hostaddr>:<port>".',
        'Multiple arguments may be given.',
        'If nothing is provided, defaults to "127.0.0.1:65000"',
    ])
    def __init__(self,
                 option_strings,
                 dest,
                 nargs=None,
                 const=None,
                 default=[('127.0.0.1', 65000)],
                 type_=str,
                 choices=None,
                 required=False,
                 help=_default_help,
                 metavar=None):
        super().__init__(
            option_strings, dest, nargs, const, default,
            type_, choices, required, help, metavar,
        )

    def __call__(self, parser, namespace, values, option_string=None):
        addr, port = values.split(':')
        values = (addr, int(port))
        items = getattr(namespace, self.dest, None)
        if items == [('127.0.0.1', 65000)]:
            items = []
        else:
            items = argparse._copy_items(items)
        items.append(values)
        setattr(namespace, self.dest, items)

def main():
    p = argparse.ArgumentParser()
    p.add_argument(
        '-c', '--client', dest='clients', action=ClientArgAction#, type=str,
    )
    args = p.parse_args()

    logger.info(f'Sending to clients: {args.clients!r}')

    loop = asyncio.get_event_loop()
    sender = UmdSender(clients=args.clients)
    loop.run_until_complete(sender.open())
    try:
        loop.run_forever()
    except KeyboardInterrupt:
        loop.run_until_complete(sender.close())
    finally:
        loop.close()

if __name__ == '__main__':
    main()



try:
    from loguru import logger
except ImportError: # pragma: no cover
    import logging
    logger = logging.getLogger(__name__)
import asyncio
from typing import Dict, Tuple, Set, Optional

from pydispatch import Dispatcher, Property, DictProperty, ListProperty

from tslumd import Tally, TallyColor, Message, Display
from tslumd.utils import logger_catch

__all__ = ('UmdReceiver',)




class UmdProtocol(asyncio.DatagramProtocol):
    def __init__(self, receiver: 'UmdReceiver'):
        self.receiver = receiver
    def connection_made(self, transport):
        logger.debug(f'transport={transport}')
        self.transport = transport
        self.receiver.connected_evt.set()
    def datagram_received(self, data, addr):
        self.receiver.parse_incoming(data, addr)


class UmdReceiver(Dispatcher):
    """Receiver for UMD messages

    Arguments:
        hostaddr (str): The local host address to bind the server to. Defaults
            to :attr:`DEFAULT_HOST`
        hostport (int): The port to listen on. Defaults to :attr:`DEFAULT_PORT`

    :Events:
        .. on_tally_added(tally: Tally)

            Fired when a :class:`~.Tally` instance is added to :attr:`tallies`

        .. on_tally_updated(tally: Tally)

            Fired when any :class:`~.Tally` instance has been updated
    """

    DEFAULT_HOST: str = '0.0.0.0' #: The default host address to listen on
    DEFAULT_PORT: int = 65000 #: The default host port to listen on

    tallies: Dict[int, Tally]
    """Mapping of :class:`~.Tally` objects using
    the :attr:`~.Tally.index` as keys
    """

    running: bool
    """``True`` if the client / server are running
    """

    loop: asyncio.BaseEventLoop
    """The :class:`asyncio.BaseEventLoop` associated with the instance"""

    _events_ = ['on_tally_added', 'on_tally_updated']
    def __init__(self, hostaddr: str = DEFAULT_HOST, hostport: int = DEFAULT_PORT):
        self.__hostaddr = hostaddr
        self.__hostport = hostport
        self.tallies = {}
        self.loop = asyncio.get_event_loop()
        self.running = False
        self._connect_lock = asyncio.Lock()
        self.connected_evt = asyncio.Event()

    @property
    def hostaddr(self) -> str:
        """The local host address to bind the server to
        """
        return self.__hostaddr

    @property
    def hostport(self) -> int:
        """The port to listen on
        """
        return self.__hostport

    async def open(self):
        """Open the server
        """
        async with self._connect_lock:
            if self.running:
                return
            logger.debug('UmdReceiver.open()')
            self.running = True
            self.connected_evt.clear()
            self.transport, self.protocol = await self.loop.create_datagram_endpoint(
                lambda: UmdProtocol(self),
                local_addr=(self.hostaddr, self.hostport),
                reuse_port=True,
            )
            await self.connected_evt.wait()
            logger.info('UmdReceiver running')

    async def close(self):
        """Close the server
        """
        async with self._connect_lock:
            if not self.running:
                return
            logger.debug('UmdReceiver.close()')
            self.running = False
            self.transport.close()
            self.connected_evt.clear()
            logger.info('UmdReceiver closed')

    async def set_bind_address(self, hostaddr: str, hostport: int):
        """Set the :attr:`hostaddr` and :attr:`hostport` and restart the server
        """
        if hostaddr == self.hostaddr and hostport == self.hostport:
            return
        running = self.running
        if running:
            await self.close()
        self.__hostaddr = hostaddr
        self.__hostport = hostport
        if running:
            await self.open()

    async def set_hostaddr(self, hostaddr: str):
        """Set the :attr:`hostaddr` and restart the server
        """
        await self.set_bind_address(hostaddr, self.hostport)

    async def set_hostport(self, hostport: int):
        """Set the :attr:`hostport` and restart the server
        """
        await self.set_bind_address(self.hostaddr, hostport)

    @logger_catch
    def parse_incoming(self, data: bytes, addr: Tuple[str, int]):
        """Parse data received by the server
        """
        while True:
            message, remaining = Message.parse(data)
            for display in message.displays:
                self.update_display(display)
            if not len(remaining):
                break

    def update_display(self, rx_display: Display):
        """Update or create a :class:`~.Tally` from data received
        by the server
        """
        if rx_display.index not in self.tallies:
            tally = Tally.from_display(rx_display)
            self.tallies[rx_display.index] = tally
            logger.debug(f'New Tally: {tally}')
            self.emit('on_tally_added', self.tallies[rx_display.index])
            return
        tally = self.tallies[rx_display.index]
        changed = tally.update_from_display(rx_display)
        if changed:
            self.emit('on_tally_updated', tally)

    async def __aenter__(self):
        await self.open()
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        await self.close()


if __name__ == '__main__':
    loop = asyncio.get_event_loop()
    umd = UmdReceiver()

    loop.run_until_complete(umd.open())
    try:
        loop.run_forever()
    except KeyboardInterrupt:
        loop.run_until_complete(umd.close())
    finally:
        loop.close()



import enum

__all__ = ('TallyColor', 'TallyType', 'TallyState')

class TallyColor(enum.IntEnum):
    """Color enum for tally indicators"""
    OFF = 0   #: Off
    RED = 1   #: Red
    GREEN = 2 #: Green
    AMBER = 3 #: Amber

class TallyType(enum.Enum):
    """Enum for the three tally display types in the UMD protocol
    """
    no_tally = 0  #: No-op
    rh_tally = 1  #: "Right-hand" tally
    txt_tally = 2 #: "Text" tally
    lh_tally = 3  #: "Left-hand" tally

class TallyState(enum.IntFlag):
    OFF = 0     #: Off
    PREVIEW = 1 #: Preview
    PROGRAM = 2 #: Program



try:
    from loguru import logger
except ImportError: # pragma: no cover
    import logging
    logger = logging.getLogger(__name__)
from typing import Dict, Set

from pydispatch import Dispatcher, Property

from tslumd import TallyType, TallyColor

__all__ = ('Tally',)

class Tally(Dispatcher):
    """A single tally object

    Properties:
        rh_tally (TallyColor): State of the "right-hand" tally indicator
        txt_tally (TallyColor): State of the "text" tally indicator
        lh_tally (TallyColor): State of the "left-hand" tally indicator
        brightness (int): Tally indicator brightness from 0 to 3
        text (str): Text to display

    :Events:
        .. event:: on_update(instance: Tally, props_changed: Sequence[str])

            Fired when any property changes

    """
    rh_tally = Property(TallyColor.OFF)
    txt_tally = Property(TallyColor.OFF)
    lh_tally = Property(TallyColor.OFF)
    brightness = Property(3)
    text = Property('')
    _events_ = ['on_update']
    _prop_attrs = ('rh_tally', 'txt_tally', 'lh_tally', 'brightness', 'text')
    def __init__(self, index_, **kwargs):
        self.__index = index_
        self._updating_props = False
        self.update(**kwargs)
        self.bind(**{prop:self._on_prop_changed for prop in self._prop_attrs})

    @property
    def index(self) -> int:
        """Index of the tally object from ``0`` to ``65534``
        """
        return self.__index

    @classmethod
    def from_display(cls, display: 'tslumd.Display') -> 'Tally':
        """Create an instance from the given :class:`~.messages.Display` object
        """
        kw = {attr:getattr(display, attr) for attr in cls._prop_attrs}
        return cls(display.index, **kw)

    def update(self, **kwargs) -> Set[str]:
        """Update any known properties from the given keyword-arguments

        Returns:
            set: The property names, if any, that changed
        """
        log_updated = kwargs.pop('LOG_UPDATED', False)
        props_changed = set()
        self._updating_props = True
        for attr in self._prop_attrs:
            if attr not in kwargs:
                continue
            val = kwargs[attr]
            if getattr(self, attr) == val:
                continue
            props_changed.add(attr)
            setattr(self, attr, kwargs[attr])
            if log_updated:
                logger.debug(f'{self!r}.{attr} = {val!r}')
        self._updating_props = False
        if len(props_changed):
            self.emit('on_update', self, props_changed)
        return props_changed

    def update_from_display(self, display: 'tslumd.messages.Display') -> Set[str]:
        """Update this instance from the values of the given
        :class:`~.messages.Display` object

        Returns:
            set: The property names, if any, that changed
        """
        kw = {attr:getattr(display, attr) for attr in self._prop_attrs}
        kw['LOG_UPDATED'] = True
        return self.update(**kw)

    def to_dict(self) -> Dict:
        """Serialize to a :class:`dict`
        """
        d = {attr:getattr(self, attr) for attr in self._prop_attrs}
        d['index'] = self.index
        return d

    # def to_display(self) -> 'tslumd.messages.Display':
    #     """Create a :class:`~.messages.Display` from this instance
    #     """
    #     kw = self.to_dict()
    #     return Display(**kw)

    def _on_prop_changed(self, instance, value, **kwargs):
        if self._updating_props:
            return
        prop = kwargs['property']
        self.emit('on_update', self, [prop.name])

    def __eq__(self, other):
        if not isinstance(other, Tally):
            return NotImplemented
        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        if not isinstance(other, Tally):
            return NotImplemented
        return self.to_dict() != other.to_dict()

    def __repr__(self):
        return f'<{self.__class__.__name__}: ({self})>'

    def __str__(self):
        return f'{self.index} - "{self.text}"'



import asyncio
import dataclasses
from dataclasses import dataclass, field
import enum
import struct
from typing import List, Tuple, Dict

from tslumd import TallyColor, Tally

__all__ = ('Display', 'Message')

class Flags(enum.IntFlag):
    """Message flags
    """
    NO_FLAGS = 0 #: No flags set
    UTF16 = 1
    """Indicates text formatted as ``UTF-16LE`` if set, otherwise ``UTF-8``"""

    SCONTROL = 2
    """Indicates the message contains ``SCONTROL`` data if set, otherwise ``DMESG``
    """

@dataclass
class Display:
    """A single tally "display"
    """
    index: int #: The display index
    rh_tally: TallyColor = TallyColor.OFF #: Right hand tally indicator
    txt_tally: TallyColor = TallyColor.OFF #: Text tally indicator
    lh_tally: TallyColor = TallyColor.OFF #: Left hand tally indicator
    brightness: int = 3 #: Display brightness (from 0 to 3)
    text: str = '' #: Text to display
    @classmethod
    def from_dmsg(cls, flags: Flags, dmsg: bytes) -> Tuple['Display', bytes]:
        """Construct an instance from a ``DMSG`` portion of received message.

        Any remaining message data after the relevant ``DMSG`` is returned along
        with the instance.
        """
        hdr = struct.unpack('<2H', dmsg[:4])
        dmsg = dmsg[4:]
        ctrl = hdr[1]
        kw = dict(
            index=hdr[0],
            rh_tally=TallyColor(ctrl & 0b11),
            txt_tally=TallyColor(ctrl >> 2 & 0b11),
            lh_tally=TallyColor(ctrl >> 4 & 0b11),
            brightness=ctrl >> 6 & 0b11,
        )
        is_control_data = ctrl & 0x8000 == 0x8000
        if is_control_data:
            raise ValueError('Control data undefined for UMDv5.0')
        else:
            txt_byte_len = struct.unpack('<H', dmsg[:2])[0]
            dmsg = dmsg[2:]
            txt_bytes = dmsg[:txt_byte_len]
            dmsg = dmsg[txt_byte_len:]
            if Flags.UTF16 in flags:
                txt = txt_bytes.decode('UTF-16le')
            else:
                if b'\0' in txt_bytes:
                    txt_bytes = txt_bytes.split(b'\0')[0]
                txt = txt_bytes.decode('UTF-8')
            kw['text'] = txt
        return cls(**kw), dmsg

    def to_dmsg(self, flags: Flags) -> bytes:
        """Build ``dmsg`` bytes to be included in a message
        (called from :meth:`Message.build_message`)
        """
        ctrl = self.rh_tally & 0b11
        ctrl += (self.txt_tally & 0b11) << 2
        ctrl += (self.lh_tally & 0b11) << 4
        ctrl += (self.brightness & 0b11) << 6
        if Flags.UTF16 in flags:
            txt_bytes = bytes(self.text, 'UTF16-le')
        else:
            txt_bytes = bytes(self.text, 'UTF-8')
        txt_byte_len = len(txt_bytes)
        data = bytearray(struct.pack('<3H', self.index, ctrl, txt_byte_len))
        data.extend(txt_bytes)
        return data

    def to_dict(self) -> Dict:
        return dataclasses.asdict(self)

    @classmethod
    def from_tally(cls, tally: Tally) -> 'Display':
        """Create a :class:`Display` from the given :class:`~.Tally`
        """
        kw = tally.to_dict()
        return cls(**kw)

    def __eq__(self, other):
        if not isinstance(other, (Display, Tally)):
            return NotImplemented
        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        if not isinstance(other, (Display, Tally)):
            return NotImplemented
        return self.to_dict() != other.to_dict()

@dataclass
class Message:
    """A single UMDv5 message packet
    """
    version: int = 0 #: Protocol minor version
    flags: int = Flags.NO_FLAGS #: The message :class:`Flags` field
    screen: int = 0 #: Screen index
    displays: List[Display] = field(default_factory=list)
    """A list of :class:`Display` instances"""

    scontrol: bytes = b''
    """SCONTROL data (if present).  Not currently implemented"""

    def __post_init__(self):
        if not isinstance(self.flags, Flags):
            self.flags = Flags(self.flags)

    @classmethod
    def parse(cls, msg: bytes) -> Tuple['Message', bytes]:
        """Parse incoming message data to create a :class:`Message` instance.

        Any remaining message data after parsing is returned along with the instance.
        """
        data = struct.unpack('<HBBH', msg[:6])
        byte_count, version, flags, screen = data
        kw = dict(
            version=version,
            flags=Flags(flags),
            screen=screen,
        )
        msg = msg[2:]
        remaining = msg[byte_count:]
        msg = msg[4:byte_count]
        obj = cls(**kw)
        if Flags.SCONTROL in obj.flags:
            obj.scontrol = msg
            return obj, remaining
        while len(msg):
            disp, msg = Display.from_dmsg(obj.flags, msg)
            obj.displays.append(disp)
        return obj, remaining

    def build_message(self) -> bytes:
        """Build a message packet from data in this instance
        """
        if Flags.SCONTROL in self.flags:
            payload = bytearray(self.scontrol)
        else:
            payload = bytearray()
            for display in self.displays:
                payload.extend(display.to_dmsg(self.flags))
        payload_byte_count = len(payload)
        fmt = f'<HBBH{payload_byte_count}B'
        pbc = struct.calcsize(fmt) - 2
        data = bytearray(struct.pack('<HBBH', pbc, self.version, self.flags, self.screen))
        data.extend(payload)
        return bytes(data)
